/*
 * Copyright (c) 2010, MegaVersiya, LTD. All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without modification, are permitted provided that the following conditions are met:
 * Redistributions of source code must retain the above copyright notice, this list of conditions and the following disclaimer.
 * Redistributions in binary form must reproduce the above copyright notice, this list of conditions and the following disclaimer in the documentation
 * and/or other materials provided with the distribution.
 * Neither the name of the MegaVersiya, LTD nor the names of its contributors may be used to endorse or promote products derived from this software 
 * without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE 
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE 
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE 
 * GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH 
 * DAMAGE.
 */

#include "c-icap.h"
#include "service.h"
#include "header.h"
#include "body.h"
#include "simple_api.h"
#include "lookup_table.h"
#include "debug.h"
#include "access.h"
#include "acl.h"
#include "../../common.h"
#include "commands.h"
#include <memcache.h>
#include <errno.h>
#undef HAVE_CONFIG_H
#include <odbx.h>
#include <regex.h>
#include <syslog.h>
#include <time.h>
#include <sys/time.h>
#ifndef REG_BASIC
#define REG_BASIC 0
#endif

#define CACHETTL	3600

#define MAX_URL_SIZE	65536
#define MAX_PAGE_SIZE (MAX_URL_SIZE - CI_MAXHOSTNAMELEN)

struct memcache *mc;

int    url_filter_init_service(ci_service_xdata_t *srv_xdata,struct ci_server_conf *server_conf);
int    url_filter_post_init_service(ci_service_xdata_t *srv_xdata,struct ci_server_conf *server_conf);
void   url_filter_close_service ();
void * url_filter_init_request_data(ci_request_t *req);
void   url_filter_release_data(void *data);
int    url_filter_check_preview(char *preview_data,int preview_data_len,ci_request_t *);
int    url_filter_io(char *wbuf,int *wlen,char *rbuf,int *rlen,int iseof,ci_request_t *req);
int    url_filter_process(ci_request_t *);

static unsigned dotip2int (char *str);
static unsigned str2int (const char *str);

static char	*MemcachedServers;
static char	*Host2CatServer;
static char	*DBEngine;
static char	*DBHost;
static char	*DBPort;
static char	*DBName;
static char	*DBUser;
static char	*DBPass;
static char	*RedirectUrl;
static char	*ClientIPHeader = "X-Client-IP";
static char	*AuthUserHeader = "X-Authenticated-User";
static char	*AuthUserPrefix;
static int	AuthUserEncoded = 1;
// Values 0 - ACCEPT, 1 - REJECT, 2 - REDIRECT
static int	DBOpenErrorPolicy = 1;	// SQLite database doesn't opened
static int	DBQueryErrorPolicy = 0;	// Error executing SQLite query
static int	UnknownRolePolicy = 1;		// Can't find role for user@ip
static int	MemcachedConnectErrorPolicy = 1;// No connection to memcached server(s)
static int	MemcachedMissPolicy = 0;	// Host category missed at memcached server(s)
static int	EmptyCategoryPolicy = 0;	// No categories for host
static int	CategoryNotFoundPolicy = 0;	// Host categories doesn't exists in roles definition
static int	SyslogPriority = LOG_INFO;	// Priority for syslog()
static char	*StatLog;			// StatLog
static int	fdStatLog;			// fd StatLog
static int	SyslogOn = 0;

static int	fd = -1;
static struct sockaddr_in	addr;
static odbx_t	*dbh;
static char	*req_type_mod;			// Type of request (RESPMOD OR REQMOD)

static ci_thread_mutex_t mtx;

static struct ci_conf_entry conf_variables[]={
     {"MemcachedServers",&MemcachedServers,ci_cfg_set_str,NULL},
     {"Host2CatServer",&Host2CatServer,ci_cfg_set_str,NULL},
     {"DBEngine",&DBEngine,ci_cfg_set_str,NULL},
     {"DBHost",&DBHost,ci_cfg_set_str,NULL},
     {"DBPort",&DBPort,ci_cfg_set_str,NULL},
     {"DBName",&DBName,ci_cfg_set_str,NULL},
     {"DBUser",&DBUser,ci_cfg_set_str,NULL},
     {"DBPass",&DBPass,ci_cfg_set_str,NULL},
     {"RedirectUrl",&RedirectUrl,ci_cfg_set_str,NULL},
     {"ClientIPHeader",&ClientIPHeader,ci_cfg_set_str,NULL},
     {"AuthUserHeader",&AuthUserHeader,ci_cfg_set_str,NULL},
     {"AuthUserPrefix",&AuthUserPrefix,ci_cfg_set_str,NULL},
     {"AuthUserEncoded",&AuthUserEncoded,ci_cfg_set_int,NULL},
     {"DBOpenErrorPolicy",&DBOpenErrorPolicy,ci_cfg_set_int,NULL},
     {"DBQueryErrorPolicy",&DBQueryErrorPolicy,ci_cfg_set_int,NULL},
     {"UnknownRolePolicy",&UnknownRolePolicy,ci_cfg_set_int,NULL},
     {"MemcachedConnectErrorPolicy",&MemcachedConnectErrorPolicy,ci_cfg_set_int,NULL},
     {"MemcachedMissPolicy",&MemcachedMissPolicy,ci_cfg_set_int,NULL},
     {"EmptyCategoryPolicy",&EmptyCategoryPolicy,ci_cfg_set_int,NULL},
     {"StatLog",&StatLog,ci_cfg_set_str,NULL},
     {"CategoryNotFoundPolicy",&CategoryNotFoundPolicy,ci_cfg_set_int,NULL},
     {"SyslogOn",&SyslogOn,ci_cfg_set_int,NULL},
     {"SyslogPriority",&SyslogPriority,ci_cfg_set_int,NULL},
     {NULL,NULL,NULL,NULL}
};


//service_module echo={
CI_DECLARE_MOD_DATA ci_service_module_t service={
     "url_filter", /*Module name*/
     "Url Filter service", /*Module short description*/
     ICAP_RESPMOD | ICAP_REQMOD, /*Service type responce or request modification*/
     url_filter_init_service, /* init_service*/
     url_filter_post_init_service, /*post_init_service*/
     url_filter_close_service, /*close_Service*/
     url_filter_init_request_data,/* init_request_data*/
     url_filter_release_data, /*Release request data*/
     url_filter_check_preview,
     url_filter_process,
     url_filter_io,
     conf_variables,
     NULL
};

int URL_FILTER_DATA_POOL = -1;
 
struct url_filter_data{
    ci_cached_file_t *body;    
    int denied;
};

/*Structs for this module */
enum http_methods {HTTP_UNKNOWN=0,HTTP_GET, HTTP_POST};

const char *protos[] = {"", "http", "https", "ftp", NULL};
enum proto {UNKNOWN=0, HTTP, HTTPS, FTP};

struct http_info{
    int http_major;
    int http_minor;
    int method;
    unsigned int port;
    int proto;
    char host[CI_MAXHOSTNAMELEN+1];
    char server_ip[64]; /*I think ipv6 address need about 32 bytes*/
    char site[CI_MAXHOSTNAMELEN+1];
    char url[MAX_URL_SIZE]; /*I think it is enough*/
    char *args;
    char *client_ip;
    char auth_user [MAX_USERNAME_LEN + 1];
};


int url_filter_init_service(ci_service_xdata_t *srv_xdata,struct ci_server_conf *server_conf){
    unsigned int xops;
    ci_debug_printf(2,"Initialization of url_filter module......\n");
    ci_service_set_preview(srv_xdata, 0);
    xops = CI_XCLIENTIP | CI_XSERVERIP;
    xops |= CI_XAUTHENTICATEDUSER | CI_XAUTHENTICATEDGROUPS;
    ci_service_set_xopts(srv_xdata, xops);
    /*initialize mempools          */
    URL_FILTER_DATA_POOL = ci_object_pool_register("url_filter_data", sizeof(struct url_filter_data));
    if (URL_FILTER_DATA_POOL < 0)
	return CI_ERROR;
    return CI_OK;
}

int url_filter_post_init_service (ci_service_xdata_t *srv_xdata,struct ci_server_conf *server_conf)
{
	int	nservers, retval;
	char	*p, *p1, *ep;

	if (!MemcachedServers || !*MemcachedServers) {
		ci_debug_printf (1, "Empty MemcachedServers list\n");
		return CI_ERROR;
	}

	if (!(mc = mc_new ())) {
		ci_debug_printf (1, "No memory for mc structure\n");
		return CI_ERROR;
	}

	ep = MemcachedServers + strlen(MemcachedServers);
	for (nservers = 0, p = MemcachedServers; p < ep; p = p1 + 1) {
		// p points to beginning of server, p1 points to end of server
		p1 = strchr (p, ',');
		if (!p1)        p1 = p + strlen (p);

		if (p1 == p) {
			ci_debug_printf (5, "Empty server in '%s' - ignored\n", MemcachedServers);
		} else if ((retval = mc_server_add5 (mc, p, p1 - p))) {
			ci_debug_printf (5, "server add for server %.*s returns %d\n", (int)(p1 - p), p, retval);
		} else {
			nservers++;
			ci_debug_printf (5, "server %.*s successfully added\n", (int)(p1 - p), p);
		}
	}

	if (!nservers) {
		ci_debug_printf (1, "No servers added\n");
		mc_free (mc);
		mc = NULL;
		return CI_ERROR;
	}

	if (!Host2CatServer || !*Host2CatServer) {
		ci_debug_printf (1, "Empty Host2CatServer\n");
		return CI_ERROR;
	}

	if (!(p = strchr (Host2CatServer, ':'))) {
		ci_debug_printf (1, "No port number in Host2CatServer %s\n", Host2CatServer);
		return CI_ERROR;
	}

	*p = 0;
	bzero (&addr, sizeof addr);
	addr.sin_family = AF_INET;
	addr.sin_port = htons (atoi (p + 1));
	inet_pton (AF_INET, Host2CatServer, &addr.sin_addr);
	*p = ':';
	fd = socket(AF_INET, SOCK_DGRAM, 0);
	if (fd == -1) {
		ci_debug_printf (1, "Can't open UDP socket to Host2CatServer: %s\n", strerror (errno));
		return CI_ERROR;
	}

	if (!DBEngine || !*DBEngine) {
		ci_debug_printf (1, "Empty DBEngine\n");
		return CI_ERROR;
	}

	retval = odbx_init (&dbh, DBEngine, DBHost, DBPort);
	if (retval != ODBX_ERR_SUCCESS) {
		ci_debug_printf (1, "Error initializing DB handle: %s\n", odbx_error (dbh, retval));
		return CI_ERROR;
	}

	if (!DBName || !*DBName) {
		ci_debug_printf (1, "Empty DBName\n");
		return CI_ERROR;
	}

	retval = odbx_bind_simple (dbh, DBName, DBUser, DBPass);
	if (retval != ODBX_ERR_SUCCESS) {
		ci_debug_printf (1, "Error binding to database: %s\n", odbx_error (dbh, retval));
		odbx_finish (dbh);
		dbh = NULL;
		return CI_ERROR;
	}

	fdStatLog = open(StatLog, O_CREAT|O_APPEND|O_WRONLY, S_IWUSR|S_IRUSR|S_IRGRP|S_IROTH);
	if (fdStatLog == -1) {
	      ci_debug_printf (1, "Can't open or create StatLog: %s\n", strerror (errno));
	      return CI_ERROR;
	}
	
	if (!RedirectUrl || !*RedirectUrl)	ci_debug_printf (1, "Empty redirect URL\n");

	if (ci_thread_mutex_init(&mtx))
		return CI_ERROR;
	return CI_OK;
}

void url_filter_close_service ()
{
	if (mc) {
		mc_free (mc);
		mc = NULL;
	}
	if (fd != -1)	close (fd);
	if (fdStatLog != -1)	close (fd);
	if (dbh) {
		odbx_unbind (dbh);
		odbx_finish (dbh);
	}
	
	ci_thread_mutex_destroy(&mtx);
	ci_object_pool_unregister(URL_FILTER_DATA_POOL);
}

void *url_filter_init_request_data(ci_request_t *req){
    struct url_filter_data *uc=ci_object_pool_alloc(URL_FILTER_DATA_POOL);
    uc->body = NULL;
    uc->denied = 0;
    return uc; /*Get from a pool of pre-allocated structs better......*/
}

void url_filter_release_data(void *data){
    struct url_filter_data *uc = data;
    if (uc->body)
	ci_cached_file_destroy(uc->body);
     ci_object_pool_free(uc);    /*Return object to pool..... */
}

int get_protocol(char *str,int size) 
{
    int i;
    for(i=0; protos[i]!=NULL; i++) {
	if(strncmp(str,protos[i],size)==0)
	    return i;
    }
    return 0;
}

static int get_http_info(ci_request_t *req,ci_headers_list_t *req_header , struct http_info *httpinf){
    char *str, *tmp;
    int i, proxy_mode=0;
    
    /*Initialize htto_info struct*/
     httpinf->url[0]='\0';
     httpinf->args = NULL;
     httpinf->site[0] = '\0';
     httpinf->host[0] = '\0';
     httpinf->server_ip[0] = '\0';
     httpinf->method = HTTP_UNKNOWN;
     httpinf->port = 0;
     httpinf->proto = UNKNOWN;
     httpinf->http_major = -1;
     httpinf->http_minor = -1;
     //httpinf->client_ip = NULL;
     //httpinf->auth_user[0] = '\0'; 

    /*Now get the site name*/
    str=ci_headers_value(req_header,"Host");
    if(str) {
    	strncpy(httpinf->site,str,CI_MAXHOSTNAMELEN);
    	httpinf->site[CI_MAXHOSTNAMELEN]='\0';
    }
     
     /*
       When x-server-ip implemented in c-icap (and squid3)
       strcpy(http->inf,req->xserverip);
       else do a getipbyname
     */
     
    str=req_header->headers[0];
    if(str[0]=='g' || str[0]=='G') /*Get request....*/
	httpinf->method=HTTP_GET;
    else if(str[0]=='p' || str[0]=='P') /*post request....*/
	httpinf->method=HTTP_POST;
     /*else unknown*/

    if((str=strchr(str,' '))==NULL){ /*The request must have the form:GETPOST page HTTP/X.X */
	 return 0;
    }
    while(*str==' ')
    	 str++;
    /*here we are at the beggining of the URL. If we are in a reqmod request the
       URL propably has the form http://site[:port]/page or just /page 
       (where, in squid transparent mode?)
     */
     /*check if we are in the form proto://url*/
     if ((tmp=strstr(str,"://"))) {	 
	proxy_mode=1;
	httpinf->proto = get_protocol(str,str-tmp);
	str = tmp+3;
	i=0;
	while(*str!=':' && *str!= '/' && i < CI_MAXHOSTNAMELEN){
		httpinf->site[i] = *str;
		httpinf->url[i] = *str;
	     	i++;
	     	str++;
	 }
	 httpinf->site[i] = '\0';
	 httpinf->url[i] = '\0';
	 if(*str==':'){
	     httpinf->port = strtol(str+1,&tmp,10);
	     if(*tmp!='/') 
		 return 0;
	     /*Do we want the port contained into URL? if no:*/
	     /*str = tmp;*/
	 }
     }     
     else {
	 strcpy(httpinf->url, httpinf->host);
	 strcpy(httpinf->site, httpinf->host);
	 httpinf->port = 80;
     }     
     i = strlen(httpinf->url);
     while (*str != ' ' && *str != '?' && *str != '\0' && i < MAX_PAGE_SIZE)    /*copy page to the struct. */
          httpinf->url[i++] = *str++;
     
     if (*str == '?') {
	  httpinf->args = &(httpinf->url[i]);
	  while (*str != ' ' && *str != '\0' && i < MAX_PAGE_SIZE) 
	      httpinf->url[i++] = *str++; /*copy page arguments*/
     }
     httpinf->url[i] = '\0';

     if (*str != ' ') {         /*Where is the protocol info????? */
          return 0;
     }
     while (*str == ' ')
          str++;
     if (*str != 'H' || *(str + 4) != '/') {    /*Not in HTTP/X.X form */
          return 0;
     }
     str += 5;
     httpinf->http_major = strtol(str, &str, 10);
     if (*str != '.') {
          return 0;
     }
     str++;
     httpinf->http_minor = strtol(str, &str, 10);

     return 1;
}

static int check_destination (struct http_info *httpinf, int policy)
{
	odbx_result_t	*sth;
	unsigned	ip, role = 0, *ncats;
	int	i, rc = CategoryNotFoundPolicy, retval;
	char	*cats = NULL, *p, *p1, buf [1024], rname[64], *uname;
	char	key[64], *val = NULL;
	int	valsz, found;
	uname = AuthUserPrefix && *AuthUserPrefix && !strncmp (httpinf->auth_user, AuthUserPrefix, strlen (AuthUserPrefix)) ? httpinf->auth_user + strlen (AuthUserPrefix) : httpinf->auth_user;
	ip = dotip2int (httpinf->client_ip);
	snprintf (key, sizeof key, "%s@%s", uname, httpinf->client_ip);
	ci_debug_printf (5, "Uname: %s\n", uname);
	ci_debug_printf (5, "IP: %s\n", httpinf->client_ip);
	ci_debug_printf (5, "URL: %s\n", httpinf->url);
	ci_debug_printf (5, "Host: %s\n", httpinf->site);

	if (!dbh) {
		ci_debug_printf (1, "DB handle doesn't opened\n");
		rc = DBOpenErrorPolicy;
		goto print_log;
	}
	// Find user role
	ci_thread_mutex_lock(&mtx);
	val = mc_aget(mc, key, strlen(key));
	ci_thread_mutex_unlock(&mtx);
	
	if (val) {
		char *ptr;
             
		ci_debug_printf(5, "Found role in cache\n");
		role = strtol(val, NULL, 10);
		ptr = strchr(val, ':') + 1;
		strncpy(rname, ptr, sizeof rname);
		free(val);
		goto cached_role;
	}

	snprintf (buf, sizeof buf, "SELECT u.addr, u.mask, u.rolid, r.rname FROM users u, roles r WHERE u.uname = '%s' AND u.rolid = r.rolid ORDER BY mask DESC", uname);

	ci_thread_mutex_lock(&mtx);
	retval = odbx_query (dbh, buf, strlen (buf));
	if (retval != ODBX_ERR_SUCCESS) {
		ci_debug_printf (1, "Error executing SQL query '%s': %s\n", buf, odbx_error (dbh, retval));
		rc = DBQueryErrorPolicy;
		ci_thread_mutex_unlock(&mtx);
		goto print_log;
	}

	retval = odbx_result (dbh, &sth, NULL, 0);
	
	if (retval < 0) {
		ci_debug_printf (1, "Error getting SQL query result '%s': %s\n", buf, odbx_error (dbh, retval));
		rc = DBQueryErrorPolicy;
		ci_thread_mutex_unlock(&mtx);
		goto print_log;
	}
	
	while ((retval = odbx_row_fetch (sth)) != 0) {
		unsigned	net, mask;
		if (retval < 0) {
			ci_debug_printf (1, "Error fetching result row '%s', :%s\n", buf, odbx_error (dbh, retval));
			break;
		}
		net = str2int (odbx_field_value (sth, 0));
		mask = str2int (odbx_field_value (sth, 1));
		if ((ip & mask) == net) {
			role = str2int (odbx_field_value (sth, 2));
			strncpy (rname, odbx_field_value (sth, 3), sizeof rname);
			break;
		}
	}

	odbx_result_finish(sth);
	ci_thread_mutex_unlock(&mtx); 

	// Find default role
	if (!role) {
		snprintf (buf, sizeof buf, "SELECT u.addr, u.mask, u.rolid, r.rname FROM users u, roles r WHERE u.uname = '%s' AND u.rolid = r.rolid ORDER BY mask DESC", "");

		ci_thread_mutex_lock(&mtx);
		retval = odbx_query (dbh, buf, strlen (buf));
		if (retval != ODBX_ERR_SUCCESS) {
			ci_debug_printf (1, "Error executing SQL query '%s': %s\n", buf, odbx_error (dbh, retval));
			rc = DBQueryErrorPolicy;
			ci_thread_mutex_unlock(&mtx);
			goto print_log;
		}

		retval = odbx_result (dbh, &sth, NULL, 0);
		if (retval < 0) {
			ci_debug_printf (1, "Error getting SQL query result '%s': %s\n", buf, odbx_error (dbh, retval));
			rc = DBQueryErrorPolicy;
			ci_thread_mutex_unlock(&mtx);
			goto print_log;
		}

		while ((retval = odbx_row_fetch (sth)) != 0) {
			unsigned	net, mask;
			if (retval < 0) {
				ci_debug_printf (1, "Error fetching result row '%s', :%s\n", buf, odbx_error (dbh, retval));
				break;
			}
			net = str2int (odbx_field_value (sth, 0));
			mask = str2int (odbx_field_value (sth, 1));
			if ((ip & mask) == net) {
				role = str2int (odbx_field_value (sth, 2));
				strncpy (rname, odbx_field_value (sth, 3), sizeof rname);
				break;
			}
		}
		odbx_result_finish(sth);
	        ci_thread_mutex_unlock(&mtx);	
	}
        
	if (role) {
		snprintf(buf, sizeof buf, "%d:%s", role, rname);
		ci_thread_mutex_lock(&mtx);
		mc_add(mc, key, strlen(key), buf, strlen(buf) + 1, CACHETTL, 0);
		ci_thread_mutex_unlock(&mtx);
	}
cached_role:
	ci_debug_printf (5, "Role: %s\n", role ? rname : "UNKNOWN");
	if (!role) {
		rc = UnknownRolePolicy;
		goto print_log;
	}

	// Read URL list
	snprintf (key, sizeof key, "url:%s", rname);
	ci_thread_mutex_lock(&mtx);
	val = mc_aget(mc, key, strlen(key));
	ci_thread_mutex_unlock(&mtx);
	if (val) {
		char *s;
		ci_debug_printf(5, "Found URL list in cache\n");

		for (s = val; *s;) {
			regex_t	preg;
			const char *url;
			int action;
			int result;

			action = strtol(s, NULL, 10);
			url = strchr(s, ':') + 1;
			s = strchr(url, ' ');
			*s++ = '\0';
		
			ci_debug_printf(9, "checking url %s against list: "
			    "action %d, regex %s", httpinf->url, action, url);
			result = regcomp(&preg, url, REG_BASIC | REG_ICASE |
			    REG_NOSUB);
			if (result) {
				ci_debug_printf(1, "Error compiling regular "
				    "expression '%s': %d\n", url, rc);
				continue;
			}
			result = regexec(&preg, httpinf->url, 0, NULL, 0);
			regfree(&preg);
			if (!result) {
				free(val);
				rc = action;
				goto print_log;
			}
		}
		free(val);
		goto cached_urls;
	}

	snprintf (buf, sizeof buf, "SELECT l.url, l.action FROM lists l, hierarchy h WHERE h.child = %d AND h.parent = l.rolid ORDER BY action DESC, rolid, stars, length (l.url)", role);

	ci_thread_mutex_lock(&mtx);
	retval = odbx_query (dbh, buf, strlen (buf));
	if (retval != ODBX_ERR_SUCCESS) {
		ci_debug_printf (1, "Error executing SQL query '%s': %s\n", buf, odbx_error (dbh, retval));
		rc = DBQueryErrorPolicy;
		ci_thread_mutex_unlock(&mtx);
		goto print_log;
	}

	retval = odbx_result (dbh, &sth, NULL, 0);
	if (retval < 0) {
		ci_debug_printf (1, "Error getting SQL query result '%s': %s\n", buf, odbx_error (dbh, retval));
		rc = DBQueryErrorPolicy;
		ci_thread_mutex_unlock(&mtx);
		goto print_log;
	}

	// Check URL list
	val = NULL;
	valsz = 0;
	found = 0;
	while ((retval = odbx_row_fetch (sth)) != 0) {
		regex_t	preg;
		const char	*url;
		int	action;
		int	result;

		if (retval < 0) {
			ci_debug_printf (1, "Error fetching result row '%s': %s\n", buf, odbx_error (dbh, retval));
			break;
		}

		if (val != (void *)-1) {
			char *ptr;
			int len;

			len = strlen(odbx_field_value(sth, 0)) +
			    strlen(odbx_field_value(sth, 1)) + 3;
			if ((ptr = realloc(val, valsz + len)) == NULL) {
				free(val);
				val = (void *)-1;
			} else {
				val = ptr;
				asprintf(&ptr, "%s:%s ",
				    odbx_field_value(sth, 1),
				    odbx_field_value(sth, 0));
				if (ptr == NULL) {
					free(val);
					val = (void *)-1;
				} else {
				        if (valsz == 0) {
						memcpy(val,ptr,len);
					} else {
						strcat(val, ptr);
					}
					valsz += len;
				}
			}
			
		}

		url = odbx_field_value (sth, 0);
		action = str2int (odbx_field_value (sth, 1));
		ci_debug_printf (9, "checking url %s against list: action %d, regex %s", httpinf->url, action, url);
		result = regcomp (&preg, url, REG_BASIC | REG_ICASE | REG_NOSUB);
		if (result) {
			ci_debug_printf (1, "Error compiling regular expression '%s': %d\n", url, rc);
			continue;
		}
		result = regexec (&preg, httpinf->url, 0, NULL, 0);
		regfree (&preg);
		if (!result && !found) {
			found = 1;
			rc = action;
		}
	}
	odbx_result_finish (sth);
        ci_thread_mutex_unlock(&mtx);
       
	if (val && val != (void *)-1) {
		ci_thread_mutex_lock(&mtx);
		mc_add(mc, key, strlen(key), val, strlen(val) + 1,
		    CACHETTL, 0);
		ci_thread_mutex_unlock(&mtx);
		free(val);
	}

	if (found)
		goto print_log;
cached_urls:
	// Get host category
	if (!mc) {
		rc = MemcachedConnectErrorPolicy;
		goto print_log;
	}
	
	ci_thread_mutex_lock(&mtx);
	cats = mc_aget (mc, (char *) httpinf->site, strlen (httpinf->site));
	ci_thread_mutex_unlock(&mtx);
	ci_debug_printf (5, "Cats: %s\n", cats ? cats : "MISSED");
	if (!cats) {
		if (fd != -1) {
			sendto (fd, httpinf->site, strlen (httpinf->site) + 1, 0, (struct sockaddr *)&addr, sizeof addr);
			struct timeval timenow;
			if (req_type_mod == "RESPMOD") {
			    usleep(500000);
			    gettimeofday(&timenow, NULL);
			    if (SyslogOn == 1)
			      syslog(SyslogPriority,"%s %u %u %s", "RESPMOD SEARCH", timenow.tv_sec, timenow.tv_usec, httpinf->url);
			    ci_thread_mutex_lock(&mtx);
			    cats = mc_aget (mc, (char *) httpinf->site, strlen (httpinf->site));
			    ci_thread_mutex_unlock(&mtx);
			    if (cats) goto continue_cached_urls;
			}
		} 
		rc = MemcachedMissPolicy;
		goto print_log;
	}
continue_cached_urls:
	if (!*cats) {
		rc = EmptyCategoryPolicy;
		goto print_log;
	}
	
	// Convert cats to array of unsigned
	for (i = 1, p = strchr (cats, ':'); p; p = strchr (p + 1, ':'), i++)
		;
	ncats = calloc (i + 1, sizeof (unsigned));
	for (i = 0, p = cats, p1 = strchr (p, ':'); *p; p1 = strchr (p, ':'), i++) {
		if (p1)	*p1 = 0;
		ncats [i] = str2int (p);
		if (p1) {
			*p1 = ':';
			p = p1 + 1;
		} else
			p += strlen (p);
	}
	ncats [i] = 0;
		
	// Read category list
	snprintf(key, sizeof key, "cat:%d", role);
	ci_thread_mutex_lock(&mtx);
	val = mc_aget(mc, key, strlen(key));
	ci_thread_mutex_unlock(&mtx);
	if (val) {
		char *s;
		ci_debug_printf(5, "Found category list in cache\n");

		for (s = val; *s;) {
			char *catstr;
			int cat, action;
			unsigned *u;

			action = strtol(s, NULL, 10);
			catstr = strchr(s, ':') + 1;
			s = strchr(catstr, ' ');
			*s++ = '\0';

			cat = str2int(catstr);
			for (u = ncats; *u; u++) {
				ci_debug_printf(9, "checking host category %d "
				    "against REJECT and REDIRECT category %d\n",
				    *u, cat);
				if (cat == *u) {
					free(val);
					rc = action;
					goto print_log;
				}
			}
		}

		free(val);
		goto print_log;
	}
	
	snprintf (buf, sizeof buf, "SELECT p.cat, p.action FROM perms p, hierarchy h WHERE h.child = %d AND h.parent = p.rolid ORDER BY action DESC, rolid", role);
        
	ci_thread_mutex_lock(&mtx);
	retval = odbx_query (dbh, buf, strlen (buf));
	if (retval != ODBX_ERR_SUCCESS) {
		ci_debug_printf (1, "Error executing SQL query '%s': %s\n", buf, odbx_error (dbh, retval));
		free (ncats);
		rc = DBQueryErrorPolicy;
		ci_thread_mutex_unlock(&mtx);
		goto print_log;
	}

	retval = odbx_result (dbh, &sth, NULL, 0);
	if (retval < 0) {
		ci_debug_printf (1, "Error getting SQL query result '%s': %s\n", buf, odbx_error (dbh, retval));
		free (ncats);
		rc = DBQueryErrorPolicy;
		ci_thread_mutex_unlock(&mtx);
		goto print_log;
	}

	// Check category list
	val = NULL;
	valsz = 0;
	found = 0;
	while ((retval = odbx_row_fetch (sth)) != 0) {
		int	cat, action;
		unsigned *u;

		if (retval < 0) {
			ci_debug_printf (1, "Error fetching result row '%s': %s\n", buf, odbx_error (dbh, retval));
			break;
		}

		if (val != (void *)-1) {
			char *ptr;
			int len;

			len = strlen(odbx_field_value(sth, 0)) +
			    strlen(odbx_field_value(sth, 1)) + 3;
			if ((ptr = realloc(val, valsz + len)) == NULL) {
				free(val);
				val = (void *)-1;
			} else {
				val = ptr;
				asprintf(&ptr, "%s:%s ",
				    odbx_field_value(sth, 1),
				    odbx_field_value(sth, 0));
				if (ptr == NULL) {
					free(val);
					val = (void *)-1;
				} else {
				        if (valsz == 0) {
						memcpy(val, ptr, len);
					} else {
						strcat(val, ptr);
					}
					valsz += len;
				}
			}
		    	  
		}

		cat = str2int (odbx_field_value (sth, 0));
		action = str2int (odbx_field_value (sth, 1));
		for (u = ncats; *u; u++) {
			ci_debug_printf (9, "checking host category %d against REJECT and REDIRECT category %d\n", *u, cat);
			if (cat == *u && !found) {
				found = 1;
				rc = action;
			}
		}
	}
	free (ncats);
	odbx_result_finish (sth);
        ci_thread_mutex_unlock(&mtx);

	if (val && val != (void *)-1) {
		ci_thread_mutex_lock(&mtx);
		mc_add(mc, key, strlen(key), val, strlen(val) + 1,
		    CACHETTL, 0);
		ci_thread_mutex_unlock(&mtx);
		
		free(val);
	}
print_log:
	// uname, ip, url, role, cats, action
	if (SyslogOn == 1)
	  syslog (SyslogPriority, "%s %s %s %s %s %s %s",
		  req_type_mod,
		  uname,
		  httpinf->client_ip,
		  httpinf->url,
		  role ? rname : "-",
		  cats ? cats : "-",
		  rc == 0 ? "ACCEPT" : rc == 1 ? "REJECT" : "REDIRECT"
	  );
	
	char	bufstr [strlen (httpinf->url) + 200];
	struct timeval tn;
	gettimeofday(&tn, NULL);
	snprintf (bufstr, sizeof bufstr, "%u %s %s %s %s %s %s %s%s", 
		  tn.tv_sec,
		  req_type_mod,
		  uname,
		  httpinf->client_ip,
		  httpinf->url,
		  role ? rname : "-",
		  cats ? cats : "-",
		  rc == 0 ? "ACCEPT" : rc == 1 ? "REJECT" : "REDIRECT","\n");
	if (write(fdStatLog, bufstr, strlen(bufstr)) == -1)
	    ci_debug_printf (1, "Can't write to StatLog: %s\n", strerror (errno));
	
	if (cats)	free (cats);
	return rc;
}

static char *reject_message="<H1>Permission denied!<H1>";
static char *redirect_message="<H1>You are redirected!<H1>";

int url_filter_check_preview(char *preview_data,int preview_data_len, ci_request_t *req){
    ci_headers_list_t* req_header;
    struct url_filter_data *uc=ci_service_data(req);
    struct http_info httpinf;
    int rc;
    char* authuser = NULL;
    int site_val;

    if (ci_req_type(req) != ICAP_RESPMOD)
	req_type_mod = "REQMOD";
    else 
	req_type_mod = "RESPMOD";

    if((req_header=ci_http_request_headers(req))==NULL) /*It is not possible but who knows .....*/
	return CI_ERROR;
    httpinf.client_ip = ci_headers_value (req->request_header, ClientIPHeader);
    httpinf.auth_user [0] = 0;

    if ( !(authuser = ci_headers_value (req->request_header, AuthUserHeader)) ) authuser = "";
    if (AuthUserEncoded)
    	ci_base64_decode (authuser, httpinf.auth_user, MAX_USERNAME_LEN);
    else
    	strncpy(httpinf.auth_user, authuser, MAX_USERNAME_LEN);

     if (!get_http_info(req, req_header, &httpinf)) /*Unknown method or something else...*/
    	     return CI_MOD_ALLOW204;

    ci_debug_printf(9, "URL  to host %s\n", httpinf.site);
    ci_debug_printf(9, "URL  page %s\n", httpinf.url);
    if  (httpinf.client_ip && httpinf.site[0] != '\0' ) rc = check_destination(&httpinf, 1); else rc = 0 ;
    ci_debug_printf (5, "Action: %s\n", rc == 0 ? "ACCEPT" : rc == 1 ? "REJECT" : "REDIRECT");

    if (!rc) {
	 /*if we are inside preview negotiation or client allow204 responces oudsite of preview then*/
	 if(preview_data || ci_req_allow204(req)) 
	      return CI_MOD_ALLOW204;
	 /*
	   Squid does not support preview of data in reqmod requests neither 204 responces outside preview
	   so we need to read all the body if exists and send it back to squid.
	   Allocate a new body for it 
	 */
	 if(ci_req_hasbody(req)){
	      int clen=ci_http_content_length(req)+100;
	      uc->body=ci_cached_file_new(clen);
	 }

    } else if (rc == 2 && RedirectUrl && *RedirectUrl) {
	 char	buf [strlen (RedirectUrl) + 20];
	 /*The URL is not a good one so....*/
	 ci_debug_printf(9,"Oh!!! we are going to redirect this site.....\n");

	 uc->denied = 1;
	 uc->body=ci_cached_file_new(strlen(redirect_message)+10);
	 ci_http_response_create(req,1,1); /*Build the responce headers*/

	 ci_http_response_add_header(req,"HTTP/1.1 302 Found");/*Send an 302 Found http responce to web client*/
	 ci_http_response_add_header(req,"Server: C-ICAP");
	 snprintf (buf, sizeof buf, "Location: %s", RedirectUrl);
	 ci_http_response_add_header(req,buf);
	 ci_http_response_add_header(req,"Content-Type: text/html");
	 ci_http_response_add_header(req,"Content-Language: en");
	 
	 ci_cached_file_write(uc->body,redirect_message,strlen(redirect_message),1);
    } else {
	 /*The URL is not a good one so....*/
	 ci_debug_printf(9,"Oh!!! we are going to deny this site.....\n");

	 uc->denied = 1;
	 uc->body=ci_cached_file_new(strlen(reject_message)+10);
	 ci_http_response_create(req,1,1); /*Build the responce headers*/

	 ci_http_response_add_header(req,"HTTP/1.1 403 Forbidden");/*Send an 403 Forbidden http responce to web client*/
	 ci_http_response_add_header(req,"Server: C-ICAP");
	 ci_http_response_add_header(req,"Content-Type: text/html");
	 ci_http_response_add_header(req,"Content-Language: en");
	 
	 ci_cached_file_write(uc->body,reject_message,strlen(reject_message),1);
    }

    unlock_data(req);
    return CI_MOD_CONTINUE;
}


int url_filter_process(ci_request_t *req){

/*
	  printf("Buffer size=%d, Data size=%d\n ",
		 ((struct membuf *)b)->bufsize,((struct membuf *)b)->endpos);
*/  
//    printf("url_filter process\n");
     return CI_MOD_DONE;     
}

int url_filter_io(char *wbuf, int *wlen, char *rbuf, int *rlen, int iseof, ci_request_t * req)
{
     int ret;
     struct url_filter_data *uc = ci_service_data(req);
     if (!uc->body)
          return CI_ERROR;

     ret = CI_OK;
     if (uc->denied == 0) {
          if (rbuf && rlen) {
               *rlen = ci_cached_file_write(uc->body, rbuf, *rlen, iseof);
               if (*rlen == CI_ERROR)
                    ret = CI_ERROR;
          }
          else if (iseof)
               ci_cached_file_write(uc->body, NULL, 0, iseof);
     }

     if (wbuf && wlen) {
          *wlen = ci_cached_file_read(uc->body, wbuf, *wlen);
          if (*wlen == CI_ERROR)
               ret = CI_ERROR;
     }

     return ret;
}

static unsigned dotip2int (char *str)
{
	unsigned	ip = 0, byte = 0;
	for (; *str; str++) {
		if (*str == '.') {
			ip <<= 8;
			ip += byte;
			byte = 0;
		} else {
			byte *= 10;
			byte += *str - '0';
		}
	}
	ip <<= 8;
	ip += byte;
	return ip;
}

static unsigned str2int (const char *str)
{
	unsigned	ip = 0;

	if (!str)	return 0;
	for (; *str; str++) {
		ip *= 10;
		ip += *str - '0';
	}
	return ip;
}
